"""Magic functions for InteractiveShell."""

from __future__ import nested_scopes

#*****************************************************************************
#       Copyright (C) 2001 Janko Hauser <jhauser@ifm.uni-kiel.de> and
#                          Fernando Prez <fperez@pizero.colorado.edu>
#
#  Distributed under the terms of the GNU Lesser General Public License (LGPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#  The full text of the LGPL is available at:
#
#                  http://www.gnu.org/copyleft/lesser.html
#*****************************************************************************

#****************************************************************************
# Modules and globals

import __release__ as Rel
__version__ = Rel.version
__date__    = Rel.date
__author__  = '%s <%s>\n%s <%s>' % \
              ( Rel.authors['Janko'] + Rel.authors['Fernando'] )
__license__ = Rel.license

# Python standard modules
import os, sys,__builtin__,inspect,pydoc,re
from pprint import pprint,pformat


# Homebrewed
from genutils import *
from Itpl import Itpl,itpl,printpl


#****************************************************************************
def on_off(tag):
    """Return an ON/OFF string for a 1/0 input."""
    tab = {0:'OFF',1:'ON'}
    return tab[tag]

class Macro:
    """Simple class to store the value of macros as strings.

    This allows us to later exec them by checking when something is an
    instance of this class."""
    
    def __init__(self,value):
        self.value = value

class Magic:
    """Magic functions for InteractiveShell.

    Shell functions which can be reached as @function_name. All magic
    functions should accept a string, which they can parse for their own
    needs. This can make some functions easier to type, eg `@cd ../`
    vs. `@cd("../")`

    ALL definitions MUST begin with the prefix magic_. The user won't need it
    at the command line, but it is is needed in the definition. """

    # class globals
    auto_status = {0:'Automagic is OFF, @ prefix IS needed for magic functions.',
                   1:'Automagic is ON, @ prefix NOT needed for magic functions.'}
    #......................................................................
    # some utility functions first

    def __init__(self):
        self.alias_table = {}

    def lsmagic(self):
        """Return a list of currently available magic functions.

        Gives a list of the bare names after mangling (['ls','cd', ...], not
        ['magic_ls','magic_cd',...]"""

        # FIXME. This needs a cleanup, in the way the magics list is built.
        
        # magics in class definition
        class_magic = lambda fn: fn.startswith('magic_') and \
                      callable(Magic.__dict__[fn])
        # in instance namespace (run-time user additions)
        inst_magic =  lambda fn: fn.startswith('magic_') and \
                     callable(self.__dict__[fn])
        # and bound magics by user (so they can access self):
        inst_bound_magic =  lambda fn: fn.startswith('magic_') and \
                           callable(self.__class__.__dict__[fn])
        magics = filter(class_magic,Magic.__dict__.keys()) + \
                 filter(inst_magic,self.__dict__.keys()) + \
                 filter(inst_bound_magic,self.__class__.__dict__.keys())
        out = []
        for fn in magics:
            out.append(fn.replace('magic_','',1))
        out.sort()
        return out
    
    def set_shell(self,shell):
        self.shell = shell
    
    def getargspec(self,obj):
        """Get the names and default values of a function's arguments.

        A tuple of four things is returned: (args, varargs, varkw, defaults).
        'args' is a list of the argument names (it may contain nested lists).
        'varargs' and 'varkw' are the names of the * and ** arguments or None.
        'defaults' is an n-tuple of the default values of the last n arguments.

        Modified version of inspect.getargspec from the Python Standard
        Library."""

        if inspect.isfunction(obj):
            func_obj = obj
        elif inspect.ismethod(obj):
            func_obj = obj.im_func
        else:
            raise TypeError, 'arg is not a Python function'
        args, varargs, varkw = inspect.getargs(func_obj.func_code)
        return args, varargs, varkw, func_obj.func_defaults

    def _get_def(self,oname,obj):
        """Return the definition header for any callable object and a success flag."""

        # There used to be a lot of fancy code here, until I realized that the
        # proper way of calling formatargspec() is with a * in the args! Now
        # this function is trivial.
        try:
            return oname + inspect.formatargspec(*self.getargspec(obj)), 1
        except:
            return 'Could not get definition header for ' + `oname` , 0

    def _ofind(self,oname):
        """Find an object in the available namespaces.

        self.ofind(oname) -> found,obj,ospace,docstring,ismagic

        Has special code to detect magic functions.
        """

        # the @ in magics isn't really part of the name
        oname = oname.strip()
        if oname.startswith('@'):
            oname = oname[1:]

        # build dict of namespaces:
        builtin = __builtin__.__dict__
        IP      = self.internal_ns
        user_config = self.user_config_ns
        user_interactive    = self.user_ns

        namespaces = { 'Python builtin':builtin,
                       'IPython internal':IP,
                       'User-defined configuration':user_config,
                       'Interactive':user_interactive,
                       }

        found = 0; obj = None;  ospace = None
        ds = None; ismagic = 0
        for nsname,ns in namespaces.items():
            try:
                obj = ns[oname]
            except KeyError:
                pass
            else:
                found = 1
                ospace = nsname
                ds = inspect.getdoc(obj)
        # try to see if it's magic
        if not found:
            try:
                obj = eval('self.magic_'+oname)
                found = 1
                ospace = 'Internal to IPython'
                ismagic = 1
                ds = inspect.getdoc(obj)
            except:
                pass
        # Play some games to try and find info about dotted objects
        # and for things like {}.get? or ''.remove? to work
        if not found:
            try:
                self.tmp_obj = eval(oname,user_interactive)
                found = 1
            except:
                try:
                    self.tmp_obj = eval(oname,builtin)
                    found = 1
                except:
                    pass
            if found:
                ds = inspect.getdoc(self.tmp_obj)
                ospace = 'Currently not defined in user session.'
                obj = self.tmp_obj
                del self.tmp_obj
        return found,obj,ospace,ds,ismagic
        
    def arg_err(self,func):
        """Print docstring if incorrect arguments were passed"""
        print 'Error in arguments:'
        print inspect.getdoc(func)


    def format_latex(self,str):
        """Format a string for latex inclusion."""

        # Characters that need to be escaped for latex:
        escape_re = re.compile(r'(%|_)',re.MULTILINE)
        # Magic command names as headers:
        cmd_name_re = re.compile(r'^(@.*?):',re.MULTILINE)
        # Magic commands 
        cmd_re = re.compile(r'(?P<cmd>@.+?\b)(?!\}\}:)',re.MULTILINE)
        # Paragraph continue
        par_re = re.compile(r'\\$',re.MULTILINE)

        str = cmd_name_re.sub(r'\n\\texttt{\\textbf{\1}}:',str)
        str = cmd_re.sub(r'\\texttt{\g<cmd>}',str)
        str = par_re.sub(r'\\\\',str)
        str = escape_re.sub(r'\\\1',str)
        #file('/home/fperez/ipython/doc/magic.tex','w').write(str)  # dbg
        return str

    def format_screen(self,str):
        """Format a string for screen printing.

        This removes some latex-type format codes."""
        # Paragraph continue
        par_re = re.compile(r'\\$',re.MULTILINE)
        str = par_re.sub('',str)
        return str

    #......................................................................
    # And now the actual magic functions

    # Functions for IPython shell work (vars,funcs, config, etc)
    def magic_lsmagic(self, parameter_s = ''):
        """List currently available magic functions."""
        print 'Available magic functions:\n@'+'  @'.join(self.lsmagic())
        print '\n' + Magic.auto_status[self.rc.automagic]
        return None
        
    def magic_magic(self, parameter_s = ''):
        """Print information about the magic function system."""

        mode = ''
        try:
            if parameter_s.split()[0] == '-latex':
                mode = 'latex'
        except:
            pass

        magic_docs = []
        for fname in self.lsmagic():
            mname = 'magic_' + fname
            for space in (Magic,self,self.__class__):
                try:
                    fn = space.__dict__[mname]
                except KeyError:
                    pass
                else:
                    break
            magic_docs.append('@%s:\n\t%s\n' %(fname,fn.__doc__))
        magic_docs = ''.join(magic_docs)

        if mode == 'latex':
            print self.format_latex(magic_docs)
            return
        else:
            magic_docs = self.format_screen(magic_docs)
        
        outmsg = """
IPython's 'magic' functions
===========================

The magic function system provides a series of functions which allow you to
control the behavior of IPython itself, plus a lot of system-type
features. All these functions are prefixed with a @ character, but parameters
are given without parentheses or quotes.

Example: typing '@cd mydir' (without the quotes) changes you working directory
to 'mydir', if it exists.

If you have 'automagic' enabled (via the command line option or with the
@automagic function), you don't need to type in the @ explicitly.

You can define your own magic functions to extend the system. See the supplied
ipythonrc and example-magic.py files for details (in your ipython
configuration directory, typically $HOME/.ipython/).

You can also define your own aliased names for magic functions. In your
ipythonrc file, placing a line like:

  execute __IP.magic_cl = __IP.magic_clear

will define @cl as a new name for @clear.

For a list of the available magic functions, use @lsmagic. For a description
of any of them, type @magic_name?.

Currently the magic system has the following functions:\n"""

        outmsg = ("%s\n%s\n\nSummary of magic functions (from @lsmagic):"
                  "\n\n@%s\n\n%s" % (outmsg,
                                     magic_docs,
                                     '  @'.join(self.lsmagic()),
                                     Magic.auto_status[self.rc.automagic] ) )

        page(outmsg,screen_lines=self.rc.screen_length)
  
    def magic_automagic(self, parameter_s = ''):
        """Make magic functions callable without having to type the initial @.
        
        Toggles on/off (when off, you must call it as @automagic, of
        course). Note that magic functions have lowest priority, so if there's
        a variable whose name collides with that of a magic fn, automagic
        won't work for that function (you get the variable instead). However,
        if you delete the variable (del var), the previously shadowed magic
        function becomes visible to automagic again."""
        
        self.rc.automagic = 1 - self.rc.automagic
        print '\n' + Magic.auto_status[self.rc.automagic]

    def magic_hist(self, parameter_s = ''):
        """Print input history (_i<n> variables), with most recent last.
        
        @hist [-n]       -> print at most 40 inputs (some may be multi-line)\\
        @hist [-n] n     -> print at most n inputs\\
        @hist [-n] n1 n2 -> print inputs between n1 and n2 (n2 not included)\\

        Each input's number <n> is shown, and is accessible as the
        automatically generated variable _i<n>.  Multi-line statements are
        printed starting at a new line for easy copy/paste.

        If option -n is used, input numbers are not printed. This is useful if
        you want to get a printout of many lines which can be directly pasted
        into a text editor.

        This feature is only available if numbered prompts are in use."""

        if not self.do_full_cache:
            print 'This feature is only available if numbered prompts are in use.'
            return
        args = parameter_s.split()
        if len(args) >= 1 and args[0] == '-n':
            args = args[1:]
            print_numbers = 0
        else:
            print_numbers = 1
        
        default_length = 40
        if len(args) == 0:
            final = self.outputcache.prompt_count
            init = max(1,final-default_length)
        elif len(args) == 1:
            final = self.outputcache.prompt_count
            init = max(1,final-int(args[0]))
        elif len(args) == 2:
            init = int(args[0])
            final = int(args[1])
        else:
            warn('@h takes 0, 1 or 2 arguments separated by spaces.')
            print self.magic_hist.__doc__
            return
        width = len(str(final))
        line_sep = ['','\n']
        for in_num in range(init,final):
            #inline = eval('self.user_ns["_i'+str(in_num)+'"]')
            inline = self.user_ns['_ih'][in_num]
            multiline = inline.count('\n') > 1
            if print_numbers:
                print str(in_num).ljust(width)+':'+ line_sep[multiline],
            print inline,

    def magic_doc(self, parameter_s=''):
        """Print the docstring for an object.

        If the given object is a class, it will print both the class and the
        constructor docstrings."""
        
        found,obj,ns,ds = self._ofind(parameter_s)[0:4]
        if not found:
            print 'Object',`parameter_s`,'not found.'
            return
        
        if type(obj) is types.ClassType:
            init_ds = inspect.getdoc(obj.__init__)
            output = itpl('Class Docstring:\n$indent(ds)\n'
                          'Constructor Docstring:\n$indent(init_ds)')
        elif type(obj) is types.InstanceType and hasattr(obj,'__call__'):
            call_ds = inspect.getdoc(obj.__call__)
            if call_ds:
                output = itpl('Class Docstring:\n$indent(ds)\n'
                              'Calling Docstring:\n$indent(call_ds)')
            else:
                output = ds
        else:
            output = ds
        if output == None:
            print 'No documentation found for',`parameter_s`
            return
        page(output)
        
    def magic_p(self, parameter_s=''):
        """Just a short alias for Python's 'print'."""
        exec 'print ' + parameter_s in self.user_ns

    def magic_profile(self, parameter_s=''):
        """Print your currently active profile."""
        if self.rc.profile:
            printpl('Current IPython profile: $self.rc.profile.')
        else:
            print 'No profile active.'

    def magic_oinfo(self, parameter_s=''):
        """Provide detailed information about an object.

        oinfo object is just a synonym for object? or ?object."""
        
        string_max = 200 # max size of strings to show (snipped if longer)
        shalf = int((string_max -5)/2)

        # full_info == 0 -> obj? , 1 -> obj??
        full_info = 0
        qmark1,oname,qmark2 = re.match('(\?*)([^?]*)(\??)',parameter_s).groups()
        oname = oname.strip()
        if qmark1 or qmark2:
            full_info = 1

        # look for the given object in all namespaces
        found,obj,ospace,ds,ismagic = self._ofind(oname)
        if ds: ds = indent(ds)

        # store output in a list which gets joined with \n at the end.
        out = []
        gotdef = 0
        if found:
            if ismagic:
                obj_type = 'Magic function'
            else:
                obj_type = type(obj).__name__
            out.append('Type:\t\t'+obj_type)

            try:
                bclass = obj.__class__
                out.append('Base Class:\t'+str(bclass))
            except: pass

            # String form, but snip if too long in ? form (full in ??)
            try:
                ostr = str(obj)
                if not full_info and len(ostr)>string_max:
                    ostr = ostr[:shalf] + ' <...> ' + ostr[-shalf:]
                out.append('String Form:\t'+ostr)
            except: pass

            out.append('Namespace:\t'+ospace)

            # Length (for strings and lists)
            try:
                length = str(len(obj))
                out.append('Length:\t\t'+length)
            except: pass

            # Filename where object was defined
            try:
                file = inspect.getabsfile(obj)
                if file.endswith('<string>'):
                    file = 'Dynamically generated function. No source code available.'
                out.append('File:\t\t'+file)
            except: pass

            # reconstruct the function definition and print it:
            defin,gotdef = self._get_def(oname,obj)
            if gotdef:
                out.append('Definition:\t'+defin)

        # Docstrings only in ? mode (unless source fails, see below)
        if ds and full_info == 0:
                out.append('Docstring:\n' + ds)

        # Original source code for any callable
        if found and full_info:
            try:
                source = inspect.getsource(obj)
                out.append('Source:\n'+source)
            except:
                if ds:
                    out.append('Docstring:\n' + ds)

        # Constructor docstring for classes
        if found and type(obj) is types.ClassType:
            # reconstruct the function definition and print it:
            try:
                init_def,gotdef = self._get_def(oname,obj.__init__)
                init_ds = inspect.getdoc(obj.__init__)
            except:
                gotdef = 0
                init_ds = None

            if gotdef or init_ds:
                out.append('\nConstructor information:')
                if gotdef:
                    out.append('Definition:\t'+ init_def)
                if init_ds:
                    out.append('Docstring:\n' + indent(init_ds))

        # Call form docstring for callable instances
        if found and type(obj) is types.InstanceType and hasattr(obj,'__call__'):
            call_ds = inspect.getdoc(obj.__call__)
            if call_ds:
                out.append('Callable:\tYes\n'
                    'Call docstring:\n' + indent(call_ds))
                
        # Finally send to printer/pager
        if found or ds:
            page('\n'.join(out),screen_lines = self.rc.screen_length)
        else:
            print 'Object',`oname`,'not currently defined.'
    # end oinfo

    def magic_source(self, parameter_s=''):
        """Show the source code for an object."""
        oname = parameter_s.strip()
        found,obj = self._ofind(oname)[0:2]
        if found:
            try:
                page(inspect.getsource(obj),screen_lines=self.rc.screen_length)
            except:
                print 'Source not found for object',`oname`
        else:
            print 'Object',`oname`,'not found.'

    def magic_file(self, parameter_s=''):
        """View the source file for an object through a pager.

        The file opens at the line the object definition begins. IPython will
        honor the environment variable PAGER if set, and otherwise will do its
        best to print the file in a convenient form."""

        oname = parameter_s.strip()
        found,obj = self._ofind(oname)[0:2]
        if found:
            try:
                sourcelines,lineno = inspect.getsourcelines(obj)
            except:
                print 'Unable to show file for object',`oname`
            else:
                page_file(inspect.getabsfile(obj),lineno)

        else:
            print 'Object',`oname`,'not found.'

    def magic_pfile(self, parameter_s=''):
        """Same as @file.

        In Python 2.2 file() is now a builtin, so with automagic on, 'file'
        doesn't work anymore. This alias is provided for convenience."""

        self.magic_file(parameter_s)

    def magic_pdef(self, parameter_s=''):
        """Print the definition header for any callable object.

        If the object is a class, print the constructor information."""
        
        oname = parameter_s.strip()
        found,obj = self._ofind(oname)[0:2]
        if found:
            if callable(obj):
                if type(obj) is types.ClassType:
                    print 'Printing class constructor information:'
                    print self._get_def(oname,obj.__init__)[0]
                elif type(obj) is types.InstanceType:
                    print self._get_def(oname,obj.__call__)[0]
                else:
                    print self._get_def(oname,obj)[0]

            else:
                print 'Object',`oname`,'is not callable.'
        else:
            print 'Object',`oname`,'not found.'
        
    def magic_who_ls(self, parameter_s=''):
        """Return a list of all interactive variables."""
        out = []
        for i in self.user_ns.keys():
            if not (i.startswith('_') or i.startswith('_i')) \
                   and not (self.internal_ns.has_key(i) or
                            self.user_config_ns.has_key(i)):
                out.append(i)
        # FIXME. The namespaces should be setup so that this kind of manual
        # kludges is unnecessary:
        if 'help' in out and \
           isinstance(self.user_ns['help'],pydoc.Helper):
            out.remove('help')
        out.sort()
        return out
        
    def magic_who(self, parameter_s=''):
        """Print all interactive variables, with some minimal formatting.

        This excludes executed names loaded through your configuration
        file and things which are internal to IPython.

        This is deliberate, as typically you may load many modules and the
        purpose of @who is to show you only what you've manually defined."""

        varlist = self.magic_who_ls()
        if not varlist:
            print 'Interactive namespace is empty.'
            return

        # if we have variables, move on...

        # stupid flushing problem: when prompts have no separators, stdout is
        # getting lost. I'm starting to think this is a python bug. I'm having
        # to force a flush with a print because even a sys.stdout.flush
        # doesn't seem to do anything!
        
        for i in varlist:
            print i+'\t',
            sys.stdout.flush()  # FIXME. Why the hell isn't this flushing???
        print # well, this does force a flush at the expense of an extra \n

    def magic_whos(self, parameter_s=''):
        """Like @who, but gives some extra information about each variable.

        For all variables, the type is printed. Additionally it prints:\\
          - For {},[],(): their length.\\
          - Everything else: a string representation, snipping their middle if
          too long."""
        
        varnames = self.magic_who_ls()
        if not varnames:
            print 'Interactive namespace is empty.'
            return

        # if we have variables, move on...

        # for these types, show len() instead of data:
        seq_types = [types.DictType,types.ListType,types.TupleType]

 # Find all variable names and types so we can figure out column sizes
        get_vars = lambda i: self.locals[i]
        type_name = lambda v: type(v).__name__
        varlist = map(get_vars,varnames)
        typelist = map(type_name,varlist)
        # column labels and # of spaces as separator
        varlabel = 'Variable'
        typelabel = 'Type'
        datalabel = 'Data/Length'
        colsep = 3
        # find the size of the columns to format the output nicely
        varwidth = max(max(map(len,varnames)), len(varlabel)) + colsep
        typewidth = max(max(map(len,typelist)), len(typelabel)) + colsep
        # table header
        print varlabel.ljust(varwidth) + typelabel.ljust(typewidth) + \
              datalabel+'\n' + '-'*(varwidth+typewidth+len(datalabel))
        # and the table itself
        for vname,var,vtype in zip(varnames,varlist,typelist):
            print itpl("$vname.ljust(varwidth)$vtype.ljust(typewidth)"),
            if vtype in seq_types:
                print len(var)
            else:
                vstr = str(var)
                if len(vstr) < 50:
                    print vstr
                else:
                    printpl('$vstr[:20]<...>$vstr[-20:]')
                
    def magic_reset(self, parameter_s=''):
        
        """Resets the namespace by removing all names defined by the user.

        Input/Output history are left around in case you need them."""

        ans = raw_input(
          "Once deleted, variables cannot be recovered. Proceed (y/n)? ")
        if not ans.lower() == 'y':
            print 'Nothing done.'
            return
        for i in self.magic_who_ls():
            del(self.locals[i])

    def magic_config(self,parameter_s=''):
        """Show IPython's internal configuration."""
        
        page('Current configuration structure:\n'+
             pformat(self.rc.dict()))

    def magic_logstart(self,parameter_s=''):
        """Start logging anywhere in a session.

        @logstart [log_name [log_mode]]

        If no name is given, it defaults to a file named 'log' in your
        IPYTHONDIR directory, in 'rotate' mode (see below).

        '@logstart name' saves to file 'name' in 'backup' mode.  It saves your
        history up to that point and then continues logging.

        @logstart takes a second optional parameter: logging mode. This can be one
        of (note that the modes are given unquoted):\\
          over: overwrite existing log.\\
          backup: rename (if exists) to name~ and start name.\\
          append: well, that says it.\\
          rotate: create rotating logs name.1~, name.2~, etc.
        """

        valid_modes = qw('over backup append rotate')
        if self.LOG:
            print 'Logging is already in place. Logfile:',self.LOG
            return

        par = parameter_s.strip()
        if not par:
            logname = self.LOGDEF
            logmode = 'rotate'  # use rotate for the auto-generated logs
        else:
            try:
                logname,logmode = par.split()
            except:
                try:
                    logname = par
                    logmode = 'backup'
                except:
                    warn('Usage: @log [log_name [log_mode]]')
                    return
        if not logmode in valid_modes:
            warn('Logging NOT activated.\n'
                 'Usage: @log [log_name [log_mode]]\n'
                 'Valid modes: '+str(valid_modes))
            return

        # If we made it this far, I think we're ok:
        print 'Activating auto-logging.'
        print 'Current session state plus future input saved to:',logname
        print 'Logging mode: ',logmode
        # put logname into rc struct as if it had been called on the command line,
        # so it ends up saved in the log header
        old_logfile = self.rc.opts.get('logfile','')  # in case we need to restore it
        self.rc.opts.logfile = logname
        self.LOGMODE = logmode  # FIXME: this should be set through a function.
        try:
            header = str(self.LOGHEAD)
            self.create_log(header,logname)
            self.logstart(header,logname)
        except:
            self.LOG = ''  # we are NOT logging, something went wrong
            self.rc.opts.logfile = old_logfile
            warn("Couldn't start log: "+str(sys.exc_info()[1]))
        else:  # log input history up to this point
            for line in self.user_ns['_ih'][1:]:
                self.logfile.write(line.rstrip()+'\n')
            self.logfile.flush()
        
    def magic_logoff(self,parameter_s=''):
        """Temporarily stop logging.

        You must have previously started logging."""
        self.switch_log(0)
        
    def magic_logon(self,parameter_s=''):
        """Restart logging.

        This function is for restarting logging which you've temporarily
        stopped with @logoff. For starting logging for the first time, you
        must use the @logstart function, which allows you to specify an
        optional log filename."""
        
        self.switch_log(1)
    
    def magic_logstate(self,parameter_s=''):
        """Print the status of the logging system."""

        self.logstate()
        
    def magic_pdb(self, parameter_s=''):
        """Toggle the calling of the pdb interactive debugger.

        When an exception is triggered, IPython can optionally call the
        interactive pdb debugger after the traceback printout. @pdb toggles
        this feature on and off."""
        
        self.shell.InteractiveTB.call_pdb = 1 - self.shell.InteractiveTB.call_pdb
        print 'Automatic pdb calling has been turned',\
              on_off(self.shell.InteractiveTB.call_pdb)

    def magic_run(self, parameter_s =''):
        """Run the named file inside IPython as a program.

        Usage:\\
          @run [-n] file [args]

        Parameters after the filename are passed as command-line arguments to
        the program (put in sys.argv). Then, control returns to IPython's
        prompt.

        The file is executed in a namespace initially consisting only of
        __name__=='__main__' and sys.argv constructed as indicated. It thus
        sees its environment as if it were being run as a stand-alone
        program. But after execution, the IPython interactive namespace gets
        updated with all variables defined in the program (except for __name__
        and sys.argv). This allows for very convenient loading of code for
        interactive work, while giving each program a 'clean sheet' to start
        in.

        If the -n option is used, __name__ is NOT set to '__main__'. This
        allows running scripts and reloading the definitions in them without
        triggering a call to testing routines which are often wrapped in an
        'if __name__=="__main__" clause. """

        # get arguments and set sys.argv for program to be run.
        args = parameter_s.split()
        if len(args) >= 1 and args[0] == '-n':
            args = args[1:]
            name = __name__
        else:
            name = '__main__'
        
        save_argv = sys.argv
        sys.argv = args

        try:
            filename = sys.argv[0]
        except IndexError:
            warn('you must provide at least a filename.')
            print '\n@run:\n',inspect.getdoc(self.magic_run)
            return
        prog_ns = {'__IPYTHON__':1,'__name__':name}
        self.safe_execfile(filename,prog_ns)
        
        # update IPython namespace with whatever we got from the program.
        del prog_ns['__IPYTHON__']
        self.user_ns.update(prog_ns)
        sys.argv = save_argv

    def magic_runlog(self, parameter_s =''):
        """Run files as logs.

        Usage:\\
          @runlog file1 file2 ...

        Run the named files (treating them as log files) in sequence inside
        the interpreter, and return to the prompt.  This is much slower than
        @run because each line is executed in a try/except block, but it
        allows running files with syntax errors in them.

        Normally IPython will guess when a file is one of its own logfiles, so
        you can typically use @run even for logs. This shorthand allows you to
        force any file to be treated as a log file."""

        for f in parameter_s.split():
            self.safe_execfile(f,self.user_ns,islog=1)

    def magic_macro(self,parameter_s = ''):
        """Define a set of input lines as a macro for future re-execution.

        Usage:\\
          @macro name n1:n2 n3:n4 ... n5 .. n6 ...

        This will define a global variable called `name` which is a string
        made of joining the slices and lines you specify (n1,n2,... numbers
        above) from your input history into a single string. This variable
        acts like an automatic function which re-executes those lines as if
        you had typed them. You just type 'name' at the prompt and the code
        executes.

        Note that the slices use the standard Python slicing notation (5:8
        means include lines numbered 5,6,7).

        For example, if your history contains:
        
          44: x=1\\
          45: y=3\\
          46: z=x+y\\
          47: print x\\
          48: a=5\\
          49: print 'x',x,'y',y\\

        you can create a macro with lines 44 through 47 (included) and line 49
        called my_macro with:

          In [51]: @macro my_macro 44:48 49

        Now, typing `my_macro` will re-execute all this code in one pass.

        The macro is a simple object which holds its value in an attribute,
        but the printing system checks for macros and executes them as code
        instead of printing them.  """

        args = parameter_s.split()
        name,ranges = args[0], args[1:]
        macro = []
        for chunk in ranges:
            macro.append(eval("self.user_ns['_ih']["+chunk+"]"))
        macro = Macro(''.join(flatten(macro)))
        self.user_ns.update({name:macro})
        print 'Macro created. To execute, simply type (without quotes):',`name`
        print 'Macro contents:'
        print macro.value,

    def magic_xmode(self,parameter_s = ''):
        """Switch modes for the exception handlers. Valid modes: Plain, Verbose.

        If called without arguments, acts as a toggle."""

        new_mode = parameter_s.strip().capitalize()
        try:
            self.InteractiveTB.set_mode(mode = new_mode)
            print 'Exception reporting mode:',self.InteractiveTB.mode
        except:
            warn('Error changing exception modes.\n' + str(sys.exc_info()[1]))
            
    def magic_colors(self,parameter_s = ''):
        """Switch color scheme for the prompts and exception handlers.

        Currently implemented schemes: NoColor, Linux, LightBG.

        Color scheme names are not case-sensitive."""
        
        new_scheme = parameter_s.strip()
        try:
            self.outputcache.set_colors(new_scheme)
        except:
            warn('Error changing prompt color schemes.\n'
                 + str(sys.exc_info()[1]))
        try:
            self.InteractiveTB.set_colors(scheme = new_scheme)
            self.SyntaxTB.set_colors(scheme = new_scheme)
        except:
            warn('Error changing exception color schemes.\n'
                 + str(sys.exc_info()[1]))
                
    # FIXME. This assumes the output cache (who handles printing of results)
    # is an available attribute named outcache. Fragile.
    def magic_Pprint(self, parameter_s=''):
        """Toggle pretty printing on/off."""
        
        self.outputcache.Pprint = 1 - self.outputcache.Pprint
        off_on = ['OFF','ON']
        print 'Pretty printing has been turned',off_on[self.outputcache.Pprint]
        
    def magic_abort(self, parameter_s=''):
        """Raise a SystemExit exception.

        In a normal IPython session this just exits IPython, but if IPython is
        running embedded inside other Python code, the SystemExit exception
        may propagate all the way up and fully exit the enclosing program."""
        raise SystemExit
        
    #......................................................................
    # Functions to implement unix shell-type things

    # Cool trick using Python's dynamic features:
    # @alias dynamically generates new functions as requested.
    def magic_alias(self, parameter_s = ''):
        """Define an alias for a system command.

        '@alias alias_name cmd' defines 'alias_name' as an alias for 'cmd'

        Then, typing '@alias_name params' will execute the system command 'cmd
        params' (from your underlying operating system).

        You can also define aliases with parameters using %s specifiers (one
        per parameter):
        
          In [1]: alias parts echo first %s second %s\\
          In [2]: @parts A B\\
          first A second B\\
          In [3]: @parts A\\
          Incorrect number of arguments: 2 expected.\\
          parts is an alias to: 'echo first %s second %s'\\

        If called with no parameters, @alias prints the current alias table."""

        par = parameter_s.strip()
        if not par:
            if self.rc.automagic:
                prechar = ''
            else:
                prechar = '@'
            print 'Alias\t\tSystem Command\n'+'-'*30
            aliases = self.alias_table.keys()
            aliases.sort()
            for alias in aliases:
                print prechar+alias+'\t\t'+self.alias_table[alias]
            return
        try:
            alias,cmd = par.split(' ',1)
        except:
            print inspect.getdoc(self.magic_alias)
            return
        nargs = cmd.count('%s')
        if nargs == 0:  # simple aliases
            fndef = itpl(
"""
def magic_${alias}(parameter_s = ''):
    '''Alias to the system command '$cmd' '''
    xsys('$cmd '+str(parameter_s))

self.magic_$alias = magic_$alias
""")
        else:  # parametric aliases
            fndef = itpl(
"""
def magic_${alias}(parameter_s = ''):
    '''Alias to the system command '$cmd' '''
    cmd = '$cmd'
    nargs = cmd.count('%s')
    args = str(parameter_s).split()

    if len(args) != nargs:
        print 'Incorrect number of arguments:',nargs,'expected.'
        print "$alias is an alias to: '$cmd'"
        return
    else:
        cmd_call = cmd % tuple(args)
        xsys(cmd_call)

self.magic_$alias = magic_$alias
""")
        try:
            exec fndef in globals(),locals()
        except:
            print self.magic_alias.__doc__
        self.alias_table.update({alias:cmd})
    # end magic_alias
        
    def magic_pwd(self, parameter_s = ''):
        """Return the current working directory path."""
        return os.getcwd()

    if os.name == 'posix':
        def magic_lf(self, parameter_s=''):
            """List (in color) things which are normal files."""
            self.magic_lc(parameter_s+' | grep ^-')

        def magic_ll(self, parameter_s=''):
            """List (in color) things which are symbolic links."""
            self.magic_lc(parameter_s+' | grep ^l')

        def magic_ld(self, parameter_s=''):
            """List (in color) things which are directories or links to directories."""
            self.magic_lc(parameter_s+' | grep /$')

        def magic_lx(self, parameter_s=''):
            """List (in color) things which are executable."""
            self.magic_lc(parameter_s+'| grep ^-..x')

    def magic_cd(self, parameter_s=''):
        """Change the current working directory.

        This command automatically maintains an internal list of directories
        you visit during your IPython session, in the variable _dh. The
        command @dhist shows this history nicely formatted.

        cd -<n> changes to the n-th directory in the directory history.

        cd - changes to the last visited directory.

        Note that !cd doesn't work for this purpose because the shell where
        !command runs is immediately discarded after executing 'command'."""
        
        ps = parameter_s
        if ps.startswith('-'):
            try:
                ps = self.user_ns['_dh'][
                    int(ps.replace('-','').strip())]
            except:
                try:
                    ps = self.user_ns['_dh'][-2]
                except:
                    print 'No previous directory to change to.'
                    return
        if ps:
            try:
                os.chdir(os.path.expanduser(ps))
            except OSError:
                print sys.exc_info()[1]
            else:
                self.user_ns['_dh'].append(os.getcwd())
                
        else:
            os.chdir(self.home_dir)
            self.user_ns['_dh'].append(os.getcwd())
        print self.user_ns['_dh'][-1]

    def magic_dhist(self, parameter_s=''):
        """Print your history of visited directories.

        @dhist       -> print full history\\
        @dhist n     -> print last n entries only\\
        @dhist n1 n2 -> print entries between n1 and n2 (n1 not included)\\

        This history is automatically maintained by the @cd command, and
        always available as the global list variable _dh. You can use @cd -<n>
        to go to directory number <n>."""

        dh = self.user_ns['_dh']
        if parameter_s:
            try:
                args = map(int,parameter_s.split())
            except:
                self.arg_err(Magic.magic_dhist)
                return
            if len(args) == 1:
                ini,fin = max(len(dh)-(args[0]),0),len(dh)
            elif len(args) == 2:
                ini,fin = args
            else:
                self.arg_err(Magic.magic_dhist)
                return
        else:
            ini,fin = 0,len(dh)
        nlprint(dh,
                header = 'Directory history (kept in _dh)',
                start=ini,stop=fin)

    def magic_env(self, parameter_s=''):
        """List environment variables."""
        # Unfortunately environ is an instance not a dictionary
        envdict = {}
        for key, value in os.environ.items():
            envdict[key] = value
        return envdict

    def magic_pushd(self, parameter_s=''):
        """Place the current dir on stack and change directory.
        
        Usage:\\
          @pushd ['dirname']

        @pushd with no arguments does a @pushd to your home directory.
        """
        if parameter_s == '': parameter_s = '~'
        if len(self.dir_stack)>0 and os.path.expanduser(parameter_s) != \
           os.path.expanduser(self.dir_stack[0]):
            try:
                self.magic_cd(parameter_s)
                self.dir_stack.insert(0,os.getcwd().replace(self.home_dir,'~'))
                self.magic_dirs()
            except:
                print 'Invalid directory'
        else:
            print 'You are already there!'

    def magic_popd(self, parameter_s=''):
        """Change to directory popped off the top of the stack.
        """
        if len (self.dir_stack) > 1:
            self.dir_stack.pop(0)
            self.magic_cd(self.dir_stack[0])
            print self.dir_stack[0]
        else:
            print "You can't remove the starting directory from the stack:",\
                  self.dir_stack

    def magic_dirs(self, parameter_s=''):
        """Return the current directory stack."""

        return self.dir_stack[:]
# end Magic
